@@ -1,4 +1,5 @@
 run () {
+  set -eo pipefail
   echo -n "Running" >&2
   printf " %q" "$@" >&2
   echo              >&2
@@ -6,26 +7,32 @@ run () {
 }
 
 cpu(){
+  set -eo pipefail
   ghc -e 'putStr System.Info.arch'
 }
 
 os(){
+  set -eo pipefail
   ghc -e 'putStr System.Info.os'
 }
 
 ghcjs_version(){
+  set -eo pipefail
   ghcjs --numeric-ghcjs-version
 }
 
 ghcjs_ghc_version(){
+  set -eo pipefail
   ghcjs --numeric-ghc-version
 }
 
 package_prefix(){
+    set -eo pipefail
     echo $1 | sed -n -e 's|^\([^-]*\)-.*-[^-]*$|\1|p'
 }
 
 package_hc(){
+    set -eo pipefail
     case $1 in
         ghc|ghc-prof) echo "ghc";;
         *) echo $1 | sed -n -e 's|^lib\([^-]*\)-.*-[^-]*$|\1|p';;
@@ -33,6 +40,7 @@ package_hc(){
 }
 
 package_ext(){
+    set -eo pipefail
     case $1 in
         # I'm told the ghc build uses these scripts, hence these special cases
         ghc) echo "dev";;
@@ -42,6 +50,7 @@ package_ext(){
 }
 
 packages_hc(){
+    set -eo pipefail
     hcs=`{ for i in ${DEB_PACKAGES}; do package_hc $i; done; } | LC_ALL=C sort -u`
     if [ `echo ${hcs} | wc -w` = 0 ]; then hcs=${DEB_DEFAULT_COMPILER}; fi
     if [ `echo ${hcs} | wc -w` != 1 ]; then echo "Multiple compilers not supported: ${hc}"; exit 1; fi
@@ -49,6 +58,7 @@ packages_hc(){
 }
 
 hc_libdir(){
+    set -eo pipefail
     case $1 in
       ghc) echo "usr/lib/haskell-packages/ghc/lib";;
       ghcjs) echo "usr/lib/ghcjs/.cabal/lib";;
@@ -57,10 +67,12 @@ hc_libdir(){
 }
 
 package_libdir(){
+    set -eo pipefail
     hc_libdir `package_hc $1`
 }
 
 hc_pkgdir(){
+    set -eo pipefail
     case $1 in
         ghc) echo "var/lib/ghc/package.conf.d";;
         ghcjs) echo "usr/lib/ghcjs/.ghcjs/`cpu`-`os`-`ghcjs_version`-`ghcjs_ghc_version`/ghcjs/package.conf.d";;
@@ -69,10 +81,12 @@ hc_pkgdir(){
 }
 
 package_pkgdir(){
+    set -eo pipefail
     hc_pkgdir `package_hc $1`
 }
 
 hc_prefix(){
+    set -eo pipefail
     case $1 in
       ghc) echo "usr";;
       ghcjs) echo "usr/lib/ghcjs";;
@@ -81,6 +95,7 @@ hc_prefix(){
 }
 
 hc_haddock(){
+    set -eo pipefail
     case $1 in
         ghc) echo "haddock";;
         ghcjs) echo "haddock-ghcjs";;
@@ -89,28 +104,33 @@ hc_haddock(){
 }
 
 hc_docdir(){
+    set -eo pipefail
     hc=$1
     pkgid=$2
     echo "usr/lib/${hc}-doc/haddock/${pkgid}/"
 }
 
 hc_htmldir(){
+    set -eo pipefail
     hc=$1
     CABAL_PACKAGE=$2
     echo "usr/share/doc/lib${hc}-${CABAL_PACKAGE}-doc/html/"
 }
 
 hc_hoogle(){
+    set -eo pipefail
     local hc
     hc=$1
     echo "/usr/lib/${hc}-doc/hoogle/"
 }
 
 strip_hash(){
+    set -eo pipefail
     echo "$1" | sed 's/-................................$//'
 }
 
 sort_uniq(){
+    set -eo pipefail
     {
         for i in "$@" ; do
             echo $i
@@ -119,6 +139,7 @@ sort_uniq(){
 }
 
 dependency(){
+    set -eo pipefail
     local package
     local version
     local next_upstream_version
@@ -129,6 +150,7 @@ dependency(){
 }
 
 ghc_pkg_field(){
+    set -eo pipefail
     hc=$1
     pkg=$2
     field=$3
@@ -136,6 +158,7 @@ ghc_pkg_field(){
 }
 
 providing_package_for_ghc(){
+    set -eo pipefail
     local package
     local dep
     local dir
@@ -154,7 +177,7 @@ providing_package_for_ghc(){
     lib=`ghc_pkg_field $hc $dep hs-libraries | grep -i ^hs-libraries |  sed -e 's|hs-libraries: *\([^ ]*\).*|\1|' `
     for dir in $dirs ; do
         if [ -e "${dir}/lib${lib}.a" ] ; then
-            package=`dpkg-query -S ${dir}/lib${lib}.a | cut -d':' -f 1` || exit $?
+            package=`dpkg-query -S ${dir}/lib${lib}.a | cut -d':' -f 1`
             continue
         fi
     done
@@ -162,6 +185,7 @@ providing_package_for_ghc(){
 }
 
 providing_package_for_ghc_prof(){
+    set -eo pipefail
     local package
     local dep
     local dir
@@ -180,7 +204,7 @@ providing_package_for_ghc_prof(){
     lib=`ghc_pkg_field $hc $dep hs-libraries | grep -i ^hs-libraries | sed -e 's|hs-libraries: *\([^ ]*\).*|\1|' `
     for dir in $dirs ; do
         if [ -e "${dir}/lib${lib}_p.a" ] ; then
-            package=`dpkg-query -S ${dir}/lib${lib}_p.a | cut -d':' -f 1` || exit $?
+            package=`dpkg-query -S ${dir}/lib${lib}_p.a | cut -d':' -f 1`
             continue
         fi
     done
@@ -188,26 +212,32 @@ providing_package_for_ghc_prof(){
 }
 
 cabal_package_ids(){
-    local config
+    set -eo pipefail
+    local pkg
     local package_ids
+    local ghcpkg
+    ghcpkg="`tmp_package_db $@`"
     until [ -z "$1" ]
     do
-      config=$1
-      package_ids="$package_ids `grep-dctrl -n -i -s Id "" $config`"
+      pkg="`basename "$1" .conf`"
+      package_ids="$package_ids `${ghcpkg} --simple-output field "${pkg}" id`"
       shift
     done
     echo $package_ids
 }
 
 cabal_depends(){
-    local config
+    set -eo pipefail
+    local pkg
     local dep
     local depends
     local final_depends
+    local ghcpkg
+    ghcpkg="`tmp_package_db $@`"
     until [ -z "$1" ]
     do
-      config=$1
-      depends="$depends `grep-dctrl -n -i -s Depends "" $config | tr "," " "`"
+      pkg="`basename "$1" .conf`"
+      depends="$depends `${ghcpkg} --simple-output field "${pkg}" depends`"
       shift
     done
     for dep in `sort_uniq $depends` ; do
@@ -223,6 +253,7 @@ cabal_depends(){
 }
 
 hashed_dependency(){
+    set -eo pipefail
     local hc
     local type
     local pkgid
@@ -241,6 +272,7 @@ hashed_dependency(){
 }
 
 depends_for_ghc(){
+    set -eo pipefail
     local dep
     local packages
     local pkgid
@@ -268,6 +300,7 @@ depends_for_ghc(){
 }
 
 depends_for_ghc_prof(){
+    set -eo pipefail
     local dep
     local packages
     local pkgid
@@ -295,6 +328,7 @@ depends_for_ghc_prof(){
 }
 
 usable_ghc_pkg() {
+    set -eo pipefail
     local ghcpkg
     local version
     if [ -x inplace/bin/ghc-pkg ]
@@ -314,6 +348,7 @@ usable_ghc_pkg() {
 }
 
 tmp_package_db() {
+    set -eo pipefail
     local ghcpkg
     ghcpkg="`usable_ghc_pkg`"
     if [ -n "${ghcpkg}" ]
@@ -322,13 +357,16 @@ tmp_package_db() {
         then
             mkdir debian/tmp-db
             cp $@ debian/tmp-db/
-            $ghcpkg --package-db debian/tmp-db/ recache
+            # Silence GHC 8.4's "ignoring (possibly broken) abi-depends field
+            # for packages" warning. See also, https://ghc.haskell.org/trac/ghc/ticket/14381
+            $ghcpkg --package-db debian/tmp-db/ recache >/dev/null
         fi
         echo "${ghcpkg} --package-db debian/tmp-db"
     fi
 }
 
 provides_for_ghc(){
+    set -eo pipefail
     local hc
     local dep
     local packages
@@ -342,6 +380,7 @@ provides_for_ghc(){
 }
 
 provides_for_ghc_prof(){
+    set -eo pipefail
     local hc
     local dep
     local packages
@@ -355,6 +394,7 @@ provides_for_ghc_prof(){
 }
 
 package_id_to_virtual_package(){
+        set -eo pipefail
         local hc
         local type
         local pkgid
@@ -365,9 +405,9 @@ package_id_to_virtual_package(){
         ghcpkg="$4"
         if [ -n "$ghcpkg" ]
         then
-            name=`${ghcpkg} --simple-output field "${pkgid}" name`
-            version=`${ghcpkg} --simple-output field "${pkgid}" version`
-            abi=`${ghcpkg} --simple-output field "${pkgid}" abi | cut -c1-5`
+            name=`${ghcpkg} --simple-output --unit-id field "${pkgid}" name`
+            version=`${ghcpkg} --simple-output --unit-id field "${pkgid}" version`
+            abi=`${ghcpkg} --simple-output --unit-id field "${pkgid}" abi | cut -c1-5`
             echo "lib${hc}-${name}-${type}-${version}-${abi}" | tr A-Z a-z
         else
             # We don't have a usable ghc-pkg, so we fall back to parsing the package id.
@@ -378,6 +418,7 @@ package_id_to_virtual_package(){
 }
 
 depends_for_hugs(){
+    set -eo pipefail
     local version
     local upstream_version
     version=`dpkg-query --showformat='${Version}' -W hugs`
@@ -386,6 +427,7 @@ depends_for_hugs(){
 }
 
 find_config_for_ghc(){
+    set -eo pipefail
     local f
     local pkg
     pkg=$1
@@ -409,6 +451,7 @@ find_config_for_ghc(){
 }
 
 clean_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + clean_recipe> "; set -x
     [ ! -x "${DEB_SETUP_BIN_NAME}" ] || run ${DEB_SETUP_BIN_NAME} clean
     run rm -rf dist dist-ghc dist-ghcjs dist-hugs ${DEB_SETUP_BIN_NAME} Setup.hi Setup.ho Setup.o .*config*
@@ -427,23 +470,58 @@ clean_recipe(){
 }
 
 make_setup_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + make_setup_recipe> "; set -x
     for setup in Setup.lhs Setup.hs
     do
       if test -e $setup
       then
         run ghc --make $setup -o ${DEB_SETUP_BIN_NAME}
-    exit 0
+        exit 0
       fi
     done
+    # Having a Setup.hs is considered good practice, but there are a few
+    # Haskell packages that don't, since cabal does not use it for build types
+    # other than 'Custom'. Find out the build type and use the corresponding
+    # standardized Setup.hs file. For more information, see
+    # https://www.haskell.org/cabal/users-guide/developing-packages.html#pkg-field-build-type
+    if grep -q '^build-type: \+Simple' ${CABAL_PACKAGE}.cabal; then
+        setup="/usr/share/haskell-devscripts/Setup-Simple.hs"
+    elif grep -q 'build-type: \+Configure' ${CABAL_PACKAGE}.cabal; then
+        setup="/usr/share/haskell-devscripts/Setup-Configure.hs"
+    elif grep -q 'build-type: \+Make' ${CABAL_PACKAGE}.cabal; then
+        setup="/usr/share/haskell-devscripts/Setup-Make.hs"
+    else
+        echo "Could not find a suitable Setup.hs file" >&2
+        exit 1
+    fi
+    run ghc --make $setup -o ${DEB_SETUP_BIN_NAME} -outputdir debian/tmp-setup-hs
     # PS4=$PS5
 }
 
 configure_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + configure_recipe> "; set -x
+
+    # dak gets all neurotic about file timestamps older than 1975
+    # new tarballs from Hackage have files with mtimes at the
+    # beginning of the epoch, so work around this pair of silliness
+    # by setting old mtimes to 1998
+    reftime=$(mktemp)
+    touch -d "1975-01-01 UTC" "${reftime}"
+    find . \! -newer "${reftime}" -exec touch -d "1998-01-01 UTC" {} \;
+    rm "${reftime}"
+
     hc=`packages_hc`
 
-    ENABLE_PROFILING=`{ for i in ${DEB_PACKAGES}; do package_ext $i | grep prof; done; } | LC_ALL=C sort -u | sed 's/prof/--enable-library-profiling/'`
+    ENABLE_PROFILING=""
+    for i in ${DEB_PACKAGES}; do
+        pkg_ext="`package_ext $i`"
+        if [ "$pkg_ext" == "prof" ]; then
+            ENABLE_PROFILING="--enable-library-profiling"
+            break
+        fi
+    done
     local GHC_OPTIONS
     for i in `dpkg-buildflags --get LDFLAGS`
     do
@@ -474,6 +552,7 @@ configure_recipe(){
 }
 
 build_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + build_recipe> "; set -x
     hc=`packages_hc`
     run ${DEB_SETUP_BIN_NAME} build --builddir=dist-${hc}
@@ -481,6 +560,7 @@ build_recipe(){
 }
 
 check_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + check_recipe> "; set -x
     hc=`packages_hc`
     run ${DEB_SETUP_BIN_NAME} test --builddir=dist-${hc} --show-details=direct
@@ -488,6 +568,7 @@ check_recipe(){
 }
 
 haddock_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + haddock_recipe> "; set -x
     hc=`packages_hc`
     haddock=`hc_haddock ${hc}`
@@ -500,6 +581,7 @@ haddock_recipe(){
 }
 
 extra_depends_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + extra_depends_recipe> "; set -x
     hc=$1
     pkg_config=$(LC_ALL=C.UTF-8 ${DEB_SETUP_BIN_NAME} register --builddir=dist-${hc} --gen-pkg-config --verbose=verbose+nowrap | sed -r -n '/^Creating package registration file: /s///p')
@@ -509,6 +591,7 @@ extra_depends_recipe(){
 }
 
 install_dev_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + install_dev_recipe> "; set -x
     PKG=$1
 
@@ -538,6 +621,7 @@ install_dev_recipe(){
 }
 
 install_prof_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + install_prof_recipe> "; set -x
     PKG=$1
     libdir=`package_libdir ${PKG}`
@@ -552,6 +636,7 @@ install_prof_recipe(){
 }
 
 install_doc_recipe(){
+    set -eo pipefail
     # local PS5=$PS4; PS4=" + install_doc_recipe> "; set -x
     PKG=$1
     hc=`package_hc ${PKG}`
@@ -573,10 +658,12 @@ install_doc_recipe(){
         # either libghc-*-dev or all the binary packages, neither of
         # which is desirable (see dh_link (1)).  So we just create a
         # (policy-compliant) symlink ourselves
-        source=`find debian/${PKG}/${htmldir} -name "*.txt"`
+        source="debian/${PKG}/${htmldir}/${CABAL_PACKAGE}.txt"
         dest=debian/${PKG}${hoogle}${PKG}.txt
-        run mkdir -p `dirname $dest`
-        run ln -rs -T $source $dest
+        if [ -f ${source} ]; then
+            run mkdir -p `dirname $dest`
+            run ln -rs -T $source $dest
+        fi
     fi
     run dh_haskell_depends -p${PKG}
     # PS4=$PS5
